import { Component, OnInit, AfterViewInit, OnDestroy } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import {  map, switchMap, filter, catchError, tap, skip, take, delay } from 'rxjs/operators'
import { Observable, fromEvent, combineLatest, of, pipe, timer, throwError, from } from 'rxjs';
import { HttpClientModule, HttpClient, HttpHeaders, HttpErrorResponse } from '@angular/common/http';

import * as forge from "node-forge";

import {BackendSelectionService} from '../backend-selection.service';
import { AuthToken, SshAuthzServer, KeyCert } from '../identity';
import { AuthorisationService } from '../authorisation.service';
import { APIServer } from '../apiserver';
import * as jwktossh from "jwk-to-ssh";


@Component({
  selector: 'app-keygen',
  templateUrl: './keygen.component.html',
  styleUrls: ['./keygen.component.css']
})
export class KeygenComponent implements OnInit, OnDestroy {

  private keygenWorker: Worker

  constructor(private route: ActivatedRoute, 
    private router: Router, 
    private authService: AuthorisationService, 
    private backendSelectionService: BackendSelectionService,
    private http: HttpClient)
  { }

  ngOnInit() {
    this.initPipelines();
  }

  ngOnDestroy() {
    //this.keygenWorker.terminate()
  }
 
  initPipelines() {

    const token$: Observable<AuthToken> = this.route.fragment.pipe(
      map((v) => this.extractToken(v)),
    );
    const key$ = from(window.crypto.subtle.generateKey(
      {
        name: "ECDSA",
        namedCurve: "P-256",
      },
      true,
      ["sign","verify"]
    )).pipe(
      tap((_) => console.log('calculating keys')),
      switchMap((v) => { return combineLatest([from(window.crypto.subtle.exportKey("jwk",v.privateKey)),from(window.crypto.subtle.exportKey("jwk",v.publicKey))]) }),
      tap((_) => console.log('converting to ssh format')),
      map(([key,pub]) => {
        return {'private': jwktossh.pack({'jwk': key, 'comment': '', 'public': false})+"\n", 'public': jwktossh.pack({'jwk': pub, 'comment': '', 'public': true})+"\n"};
      }),
      tap((v) => console.log('keygen complete',v)),
    )
    const apiserver$: Observable<APIServer> = this.backendSelectionService.apiserver.pipe(
      filter((v) => v !== undefined),filter((v) => v !== null),
    );

    let keycert$ = combineLatest([token$, key$, apiserver$]).pipe(
      switchMap(([token,key,apiserver]) => this.getCert(token,key,apiserver),
        ([token,key,apiserver],cert) => [key,cert,token]),
      tap(([key,cert,token]) => this.logout(token.sshauthzservice)),
    );

    let agent$ = combineLatest([keycert$.pipe(filter((v) => v !== null)),apiserver$]).pipe(
      switchMap(([keycert,apiserver]) => this.addCert(keycert,apiserver)),
      switchMap((_) => this.authService.updateAgentContents()),
      switchMap((_) => this.authService.loggedInAuthZ),
      switchMap((_) => of([null])),
      catchError((e) => { console.log('error in agent$ pipeline',e); return throwError(e)}),
      catchError((e) => { console.log('pipeline stopping',e); return of([null])})
    );
    agent$.subscribe( (res) => this.router.navigate([sessionStorage.getItem('path')]),
                      (err) => { console.log('error',err); this.logout(err.sshauthzservice); this.router.navigate(['/noaccount',err.sshauthzservice.name])})
  }

 extractToken(frag: string) {
   if (frag === undefined || frag == null) {
       return;
   }
   let tokenmatch = null;
   let statematch = null;
   if (!(frag === undefined) && !(frag == null)) {
     tokenmatch = frag.match(/access_token\=([^&]+)(&|$)/);
     statematch = frag.match(/state\=([^&]+)(&|$)/);
   }
   if (tokenmatch == null || statematch == null) {
     throw new Error('no token present');
   }

   let accesstoken = tokenmatch[1];
   let state = statematch[1];
   let tuple = JSON.parse(sessionStorage.getItem('authservice'));
   if (tuple[1] != state) {
     throw new Error('callback state parameter does not match');
   }

   return new AuthToken(tokenmatch[1],tuple[0]);

 }

  logout(sshauthzserver) {
    console.log('sshauthz server is',sshauthzserver);
   if (!(sshauthzserver.logout === null)) {
     window.open(sshauthzserver.logout);
   }
  }

  getCert(token: AuthToken, key: any, apiserver: APIServer): Observable<any> {
    let headers = new HttpHeaders();
    let options = { headers: headers, withCredentials: false};
    let data = {'token': token.token, 'pubkey': key.public, 'signing_url': token.sshauthzservice.sign};
    return this.http.post<any>(apiserver.tes+'/getcert',data, options).pipe(
      map((v) => v.cert),
      catchError((e) => { console.log('getCert error Handler'); console.error(e); console.log(token); return throwError(token) })
    )
  }

  addCert(kclist: any, apiserver: APIServer): Observable<any> {
    let keyCert = new KeyCert()
    console.log('in addCert');
    keyCert.key = kclist[0].private
    keyCert.cert = kclist[1]
    return this.sshAdd(keyCert,apiserver);
    console.log('add cert complete');
  }

 public sshAdd(keyCert: KeyCert, apiserver) {
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   var anyvar: any;
   let data = {'key': keyCert.key, 'cert': keyCert.cert};
   console.log('posting to sshagent with data',data);
   return this.http.post<any>(apiserver.tes+'/sshagent',data,options)
 }

  storeKey(keyCert: KeyCert) {
    var keys: KeyCert[] = [];
    try{
      keys = JSON.parse(sessionStorage.getItem('keys'));
    } catch {
      keys = [];
    }
    if (keys === null) {
      keys = [];
    }
    keys.push(keyCert);
    sessionStorage.setItem('keys',JSON.stringify(keys))
  }


    /*  ngAfterViewInit() {
    this.authService.getCert();
  }

  genKey() {
   let newkeypair = keypair();
   let publicKey = forge.pki.publicKeyFromPem(newkeypair.public);
   let sshpub = forge.ssh.publicKeyToOpenSSH(publicKey, 'sv2@monash.edu');
   return sshpub
  }*/

    /*  navigate(readyToNavigate: [Boolean,string]) {
    if (readyToNavigate[0]) {
      this.router.navigate([readyToNavigate[1]]);
    }
  }*/

  }
